require 'multi_json'
require 'timeout'

module BeanstalkFarmer
  class Job
    include BeanstalkFarmer

    attr_accessor :name, :args, :job, :response

    # @param [Beanstalk::Job] job A Beanstalk job that has been reserved to be
    #   worked upon
    def initialize(job)
      self.job = job
      set_name_and_arguments
      logger.info "JOB (#{name}) reserved"
    end

    # @return [#call] a class used to process the reserved job
    def handler
      self.class.handler_pool[name]
    end

    # Performs work for this job
    def work
      do_some_work
    rescue Timeout::Error
      handle_timeout!
    ensure
      handle_job_response
    end

    def method_missing(name, *args, &block)
      if job.respond_to?(name)
        meth = job.method(name)

        if meth.arity > 0
          meth.call(args, &block)
        else
          meth.call
        end
      else
        super
      end
    end

    # A pool of job handlers that can work on jobs in our queue
    #
    # @param [Boolean] reset (false) When true, our collection of job handlers
    #   will be purged from memory
    #
    # @return [Hash] the collection of job handlers
    def self.handler_pool(reset=false)
      @pool   = nil if reset
      @pool ||= Hash.new
    end

    private

    def set_name_and_arguments
      self.name, self.args = MultiJson.decode(job.body)
    end

    def do_some_work
      Timeout.timeout(job.ttr) do
        logger.info "JOB (#{name}) working"
        self.response = handler.call(args)
      end
    end

    def handle_timeout!
      logger.error "JOB (#{name}) out of time"
      raise TimedOut, "#{name} could not finish in #{job.ttr} seconds"
    end

    def handle_job_response
      if response.is_a?(Array) && !response.empty?
        logger.info "JOB (#{name}) done. Response: #{response.inspect}"
        job.public_send(*response)
      else
        logger.info "JOB (#{name}) done. Deleting by default."
        delete
      end
    rescue NoMethodError => e
      logger.info "JOB (#{name}) done. Assuming delete from response: #{response.inspect}"
      delete
    end
  end
end
