module BeanstalkFarmer
  # Maps tube names to job handlers, and manages the run loop.
  #
  # @example
  #
  #   runner = BeanstalkFarmer::Runner.new
  #
  #   runner.register_handlers do
  #     tube 'my_tube', lambda { |args| puts args {
  #   end
  #
  #   runner.run!
  class Runner
    attr_reader :handler_pool

    def initialize
      @handler_pool = Job.handler_pool
    end

    # @return [Farmer::BeanstalkService] a connection to the Beanstalk queue
    def service
      @service ||= Service.new
    end

    # Prepares tubes for watching
    def prep_tubes
      service.prep(handler_pool.keys)
    end

    # Reserves a job, and works it
    #
    # @param [Integer] timeout (nil) The optional timeout to use when
    #   reserving a job
    def reserve_and_work_job(timeout=nil)
      service.reserve_and_work_job(timeout)
    end

    # Runs a loop looking for jobs to reserve and run
    def work_jobs
      service.run_loop { reserve_and_work_job }
    end

    # Closes the connection to the Beanstalk queue
    def close_connection
      service.close
    end

    # @yield block A DSL to define jobs for your queue
    #
    # @example
    #   runner = Farmer::Runner.new
    #
    #   runner.register_handlers do
    #     tube 'sms',   proc { |args| puts args }
    #     tube 'mine',  Proc.new { |args| puts args }
    #   end
    def register_handlers(&block)
      dsl = DSL.new(handler_pool)
      dsl.eval(&block)
      dsl
    end

    # Looks for jobs to reserve, and applies handlers to them
    def run!
      prep_tubes
      work_jobs
    end
  end
end
