require 'beanstalk-client'
require 'multi_json'

module BeanstalkFarmer
  ##
  # Provides an abstraction against our Beanstalk client to buffer us against
  # changes in their API.
  #
  # Set the host and port through the configuration interface.
  #
  # @todo Abstract all 3rd party errors
  # @todo Catch all 3rd party errors, and re-raise our own
  #
  # @private
  class Service
    include BeanstalkFarmer

    DEFAULT_DELAY    = 0
    DEFAULT_PRIORITY = 65536
    DEFAULT_TTR      = 120

    attr_accessor :host, :port

    # Sets the host and port of your Beanstalk queue.
    #
    # @param [String] host (DEFAULT_HOST) the host name to of your Beanstalk queue
    #
    # @param [String] port (DEFAULT_PORT) the port that your Beanstalk queue is on
    def initialize
      self.host = Config.host
      self.port = Config.port
    end

    # @return [String] a formatted URI for a Beanstalk queue
    def uri
      build_uri(host, port)
    end

    # @return [Beanstalk::Pool] a connection to Beanstalk
    def connection
      @connection ||= Beanstalk::Pool.new([uri])
    rescue Beanstalk::NotConnected
      raise NotConnectedError
    end

    # Closes the Beanstalk connection
    #
    # @return [nil] Nothing. Absolutely nothing.
    def close
      logger.warn "Closing connection to Beanstalk"
      connection.close
      delete_connection
    end

    # @param [Array<String>] tube_names The tube names to be watched
    def prep(tube_names)
      watch_tubes(tube_names)
      ignore_unwatched_tubes(tube_names)
    end

    # @param [Array<String>] tube_names The tube names to be watched
    def watch_tubes(tube_names)
      logger.info "Watching tubes: #{tube_names.inspect}"
      tube_names.each { |tube_name| connection.watch(tube_name) }
    end

    # Ignores any tubes that aren't of interest, excluding the default tube.
    def ignore_unwatched_tubes(watched_tube_names)
      connection.list_tubes_watched.values.flatten.each do |tube_name|
        unless watched_tube_names.include?(tube_name)
          logger.info "Ignoring tube: #{tube_name}"
          connection.ignore(tube_name)
        end
      end
    end

    # Helper method. Used when testing to see if our queue works
    def enqueue(tube_name, args={})
      encoded_message = MultiJson.encode([tube_name, args])

      connection.on_tube(tube_name) do |conn|
        conn.put(encoded_message, DEFAULT_PRIORITY, DEFAULT_DELAY, DEFAULT_TTR)
      end
    end

    # @param [Integer, nil] timeout When nil, a job is reserved indefinitely,
    #   otherwise for the number of seconds provided
    #
    # @return [Farmer::Job] a Farmer Job to work
    def reserve(timeout=nil)
      Job.new(connection.reserve(timeout))
    rescue Beanstalk::TimedOut => e
      raise TimedOut, e.message
    end

    # @see BeanstalkFarmer::Service#reserve
    def reserve_and_work_job(timeout=nil)
      job = reserve(timeout)
      job.work
    end

    # @yield A block to execute in the run loop
    def run_loop(&block)
      loop &block
    end

    private

    def build_uri(host, port)
      [host, port].join(':')
    end

    def delete_connection
      @connection = nil
    end
  end
end
